#include "cdfur.h"
#include <stdlib.h>
#include <time.h>

#define CDFUR_EPSILON (0.00005f)

//---------------------------------------------------------------------------------------
CDFur::CDFur()
{
   mDebugMode = 0;
   mRandomSeed = ( unsigned int )time( NULL );
   mRandomPlacement = 1.0f;
   mRandomDirection = 0.2f;
   mStrandsPerUnit = 40;
   mStrandThickness = 0.01;
   mRandomStrandThickness = 0.0f;
   mSegments = 3;
   mRotateStrands = true;
   mLength = 0.5f;
   mRootDepth = 0.01f;
   mRandomLength = 0.0f;
   mSHead = NULL;
   mSTail = NULL;
   mGravity.x =  0.0f;
   mGravity.y = -0.2f;
   mGravity.z =  0.0f;
   mStrandSides = 1;
   mRemoveDuplicates = true;
}
//---------------------------------------------------------------------------------------
CDFur::~CDFur()
{
   // Remove the strands
   while( mSHead != NULL )
   {
      mSTail = mSHead;
      mSHead = mSHead->mNext;
      delete mSTail;
   }
}
//---------------------------------------------------------------------------------------
ACObject *CDFur::Generate( bool fur_add_undoable )
{

   if( mDebugMode ) DebugPrintSettings();

   // Grab a list of selected surfaces
   if( mDebugMode ) printf( "Gathering list of surfaces\n" );
   List *selsurf = ac_selection_get_whole_surfaces_all();
   if( selsurf == NULL )
   {
      message_dialog( "fur : no whole surfaces were selected." );
      return NULL;
   }

   // Seed the randomizer
   srand( this->mRandomSeed );

   if( mDebugMode ) printf( "Applying fur to %d surfaces\n", list_count( selsurf ));

   // Apply hair to each surface
   int surfctr = 0;
   unsigned int strandct = 0;
   for( List *sp = selsurf; sp != NULL; sp = sp->next )
   {

      if( mDebugMode ) printf( "\nProcessing surface %d\n", ++surfctr );

      Surface *currsurf = ( Surface * )( sp->data );
      int currsurfsmooth = surface_get_shading( currsurf );

      // Process each triangle
      int trictr = 0;
      List *trisurf = surface_get_triangulations( currsurf );
      for( List *stp = trisurf; stp != NULL; stp = stp->next )
      {

         if( mDebugMode > 12 ) printf( "\nProcessing triangle %d\n", ++trictr );

         Surface *trisurf = ( Surface * )( stp->data );
         SVertex *v1 = ( SVertex * )( trisurf->vertlist->data );
         SVertex *v2 = ( SVertex * )( trisurf->vertlist->next->data );
         SVertex *v3 = ( SVertex * )( trisurf->vertlist->next->next->data );

         // Calculate the area of the triangle
         float pararea = C3d::AreaOfParallelogram( v1->v, v2->v, v3->v );
         if( mDebugMode > 12 ) printf( "Calculated triangle area as %g\n", pararea );

         // Get the number of hairs for this surface - apply at least 1 hair
         float numhairs = max( 1.0f, pararea * pow( mStrandsPerUnit, 2 ));
         if( mDebugMode > 12 ) printf( "Number of hairs for this surface = %g\n", numhairs );

         // Get the dominant plane
         int domplane = C3d::GetDominantPlaneFromNormal( trisurf->normal );
         if( mDebugMode > 12 ) printf( "Dominant plane = %d\n", domplane );

         // Get the min/max width/height of the 2d projected triangle
         Point2 trimin, trimax;
         Vertex raystart;
         Vertex raydirection;
         raydirection.x = 0;
         raydirection.y = 0;
         raydirection.z = 0;
         float *raystartx, *raystarty;
         switch( domplane )
         {
         case C3D_PLANE_X:
            trimin.x = min( min( v1->v->z, v2->v->z ), v3->v->z );
            trimin.y = min( min( v1->v->y, v2->v->y ), v3->v->y );
            trimax.x = max( max( v1->v->z, v2->v->z ), v3->v->z );
            trimax.y = max( max( v1->v->y, v2->v->y ), v3->v->y );
            raydirection.x = 1.0f;
            raystart.x = min( min( v1->v->x, v2->v->x ), v3->v->x ) - 1.0f;
            raystartx = &( raystart.z );
            raystarty = &( raystart.y );
            break;
         case C3D_PLANE_Y:
            trimin.x = min( min( v1->v->x, v2->v->x ), v3->v->x );
            trimin.y = min( min( v1->v->z, v2->v->z ), v3->v->z );
            trimax.x = max( max( v1->v->x, v2->v->x ), v3->v->x );
            trimax.y = max( max( v1->v->z, v2->v->z ), v3->v->z );
            raydirection.y = 1.0f;
            raystart.y = min( min( v1->v->y, v2->v->y ), v3->v->y ) - 1.0f;
            raystartx = &( raystart.x );
            raystarty = &( raystart.z );
            break;
         case C3D_PLANE_Z:
            trimin.x = min( min( v1->v->x, v2->v->x ), v3->v->x );
            trimin.y = min( min( v1->v->y, v2->v->y ), v3->v->y );
            trimax.x = max( max( v1->v->x, v2->v->x ), v3->v->x );
            trimax.y = max( max( v1->v->y, v2->v->y ), v3->v->y );
            raydirection.z = 1.0f;
            raystart.z = min( min( v1->v->z, v2->v->z ), v3->v->z ) - 1.0f;
            raystartx = &( raystart.x );
            raystarty = &( raystart.y );
            break;
         }
         if( mDebugMode > 12 ) printf( "2D triangle area = %g:%g %g:%g\n", trimin.x, trimax.x, trimin.y, trimax.y );

         // Number of units between hairs
         float step = max( CDFUR_EPSILON, 1.0f / mStrandsPerUnit );
         if( mDebugMode > 12 ) printf( "Distance between strands = %g\n", step );

         // Adjust the x/y range to get global, uniform hair placement
         float f = ceil( trimin.x );
         while( f >= trimin.x )
         {
            f -= step;
         }
         trimin.x = f;
         f = ceil( trimin.y );
         while( f >= trimin.y )
         {
            f -= step;
         }
         trimin.y = f;
         f = floor( trimax.x );
         while( f <= trimax.x )
         {
            f += step;
         }
         trimax.x = f;
         f = floor( trimax.y );
         while( f <= trimax.y )
         {
            f += step;
         }
         trimax.y = f;
         if( mDebugMode > 12 ) printf( "Adjusted 2D triangle area = %g:%g %g:%g\n", trimin.x, trimax.x, trimin.y, trimax.y );

         // Number of hairs to apply horizontally/vertically
         float sqarea = ( trimax.x - trimin.x ) + ( trimax.y - trimin.y );
         int numhairsx = ( int )(( trimax.x - trimin.x ) / step + 1 );
         int numhairsy = ( int )(( trimax.y - trimin.y ) / step + 1 );
         if( mDebugMode > 12 ) printf( "X/Y collision attempts = %d:%d\n", numhairsx, numhairsy );

         // Create a local triangle with just vertices
         Vertex tri[3];
         tri[0].x = v1->v->x;
         tri[0].y = v1->v->y;
         tri[0].z = v1->v->z;
         tri[1].x = v2->v->x;
         tri[1].y = v2->v->y;
         tri[1].z = v2->v->z;
         tri[2].x = v3->v->x;
         tri[2].y = v3->v->y;
         tri[2].z = v3->v->z;
         if( mDebugMode > 12 ) printf( "Local triangle = %g %g %g : %g %g %g : %g %g %g\n",
                                        tri[0].x = v1->v->x, tri[0].y = v1->v->y, tri[0].z = v1->v->z,
                                        tri[1].x = v2->v->x, tri[1].y = v2->v->y, tri[1].z = v2->v->z,
                                        tri[2].x = v3->v->x, tri[2].y = v3->v->y, tri[2].z = v3->v->z );

         // Cast numhairsx * numhairsy rays through the square area of the projected triangle,
         // applying a hair where a hit is encountered.
         Vertex ipoint;
         *raystartx = trimin.x;
         Vertex dirvec;
         for( int ix = 0; ix <= numhairsx; ++ix )
         {
            *raystarty = trimin.y;
            for( int iy = 0; iy <= numhairsy; ++iy )
            {
               *raystarty += step;
               //printf( "Cast from %g %g %g\n", raystart.x, raystart.y, raystart.z );
               if( mDebugMode > 12 ) printf( "Casting ray through triangle\n" );
               int res = C3d::RayTriangleIntersect( &raystart, &raydirection, tri, &ipoint );
               if( res == C3D_ISECT_SUCCESS )
               {
                  if( mDebugMode > 12 ) printf( "Collision at %g %g %g\n", ipoint.x, ipoint.y, ipoint.z );
                  if( mRemoveDuplicates )
                  {
                     // Do not add if another strand is at this location
                     if( FindDuplicateStrand( v1->v, v2->v, v3->v, &ipoint ) != NULL )
                     {
                        if( mDebugMode > 12 ) printf( "Duplicate strand exists - not adding\n" );
                        continue;
                     }
                  }
                  ++strandct;
                  if( strandct > this->mMaxStrands )
                  {
                     break;
                  }
                  // Add a hair at this location.
                  CDStrand *strand = new CDStrand( this->mSegments );
                  strand->mParentSurface = currsurf;
                  strand->mRoot.x = ipoint.x;
                  strand->mRoot.y = ipoint.y;
                  strand->mRoot.z = ipoint.z;
                  // Randomize the placement of this strand.
                  if( this->mRandomPlacement > CDFUR_EPSILON )
                  {
                     float oldstartx = *raystartx;
                     float oldstarty = *raystarty;
                     // Randomize placement to within 50% of the original hair and the adjacent hair
                     *raystartx += (( float )(( rand() % 100 ) - 50 )) * 0.01f * step * this->mRandomPlacement;
                     *raystarty += (( float )(( rand() % 100 ) - 50 )) * 0.01f * step * this->mRandomPlacement;
                     // This intersection may be outside the triangle, but may or may not still be on a valid surface.
                     C3d::RayTriangleIntersect( &raystart, &raydirection, tri, &ipoint );
                     *raystartx = oldstartx;
                     *raystarty = oldstarty;
                  }
                  if( currsurfsmooth == 0 )
                  {
                     // Uniform normal for flat surface
                     strand->mDirection.x = trisurf->normal.x;
                     strand->mDirection.y = trisurf->normal.y;
                     strand->mDirection.z = trisurf->normal.z;
                  }
                  else
                  {
                     // Calculate normal based on location on surface
                     C3d::InterpolateNormal( ipoint, v1, v2, v3, strand->mDirection );
                  }
                  // Apply random direction
                  if( this->mRandomDirection > CDFUR_EPSILON )
                  {
                     strand->mDirection.x += max( -1.0f, min( 1.0f, (( float )(( rand() % 100 ) - 50 )) * 0.01f * this->mRandomDirection ));
                     strand->mDirection.y += max( -1.0f, min( 1.0f, (( float )(( rand() % 100 ) - 50 )) * 0.01f * this->mRandomDirection ));
                     strand->mDirection.z += max( -1.0f, min( 1.0f, (( float )(( rand() % 100 ) - 50 )) * 0.01f * this->mRandomDirection ));
                  }
                  // Normalize the direction vector to 1
                  float veclen = pow( pow( strand->mDirection.x, 2 ) + pow( strand->mDirection.y, 2 ) + pow( strand->mDirection.z, 2 ), 0.5f );
                  if( veclen != 0 )
                  {
                     strand->mDirection.x /= veclen;
                     strand->mDirection.y /= veclen;
                     strand->mDirection.z /= veclen;
                  }
                  // Store a copy of the direction vector
                  dirvec.x = strand->mDirection.x;
                  dirvec.y = strand->mDirection.y;
                  dirvec.z = strand->mDirection.z;
                  // Embed strand under surface
                  if( fabs( mRootDepth ) > CDFUR_EPSILON )
                  {
                     ipoint.x -= mRootDepth * dirvec.x;
                     ipoint.y -= mRootDepth * dirvec.y;
                     ipoint.z -= mRootDepth * dirvec.z;
                  }
                  // Get randomized length
                  float slenrand = 0.0f;
                  if( this->mRandomLength > CDFUR_EPSILON )
                  {
                     // Vary length from -100% to 100%
                     slenrand = (( float )( rand() % 2000 - 1000 )) * 0.001f * this->mRandomLength * this->mLength;
                  }
                  // Add the segments
                  float seglength = ( mLength + slenrand ) / mSegments;
                  for( int ix = 0; ix < mSegments + 1; ++ix )
                  {
                     if( ix == 0 )
                     {
                        strand->mSegmentVerts[ix].x = ipoint.x;
                        strand->mSegmentVerts[ix].y = ipoint.y;
                        strand->mSegmentVerts[ix].z = ipoint.z;
                     }
                     else
                     {
                        // Add gravity to each segment
                        dirvec.x += this->mGravity.x;
                        dirvec.y += this->mGravity.y;
                        dirvec.z += this->mGravity.z;
                        // Normalize the direction vector to 1 for calculations, but do not store back to dirvec
                        float veclen = pow( pow( dirvec.x, 2 ) + pow( dirvec.y, 2 ) + pow( dirvec.z, 2 ), 0.5f );
                        if( veclen != 0 )
                        {
                           strand->mSegmentDirs[ix-1].x = dirvec.x / veclen;
                           strand->mSegmentDirs[ix-1].y = dirvec.y / veclen;
                           strand->mSegmentDirs[ix-1].z = dirvec.z / veclen;
                        }
                        strand->mSegmentVerts[ix].x = strand->mSegmentVerts[ix-1].x + strand->mSegmentDirs[ix-1].x * seglength;
                        strand->mSegmentVerts[ix].y = strand->mSegmentVerts[ix-1].y + strand->mSegmentDirs[ix-1].y * seglength;
                        strand->mSegmentVerts[ix].z = strand->mSegmentVerts[ix-1].z + strand->mSegmentDirs[ix-1].z * seglength;
                        // Assign final direction
                        if( ix == mSegments - 1 )
                        {
                           strand->mSegmentDirs[ix].x = strand->mSegmentDirs[ix-1].x;
                           strand->mSegmentDirs[ix].y = strand->mSegmentDirs[ix-1].y;
                           strand->mSegmentDirs[ix].z = strand->mSegmentDirs[ix-1].z;
                        }
                     }
                  }
                  if( mSHead == NULL )
                  {
                     mSHead = strand;
                     mSTail = mSHead;
                  }
                  else
                  {
                     mSTail->mNext = strand;
                     mSTail = mSTail->mNext;
                  }
                  //printf( "Found intersection at %g %g %g\n", ipoint.x, ipoint.y, ipoint.z );
               }
            }
            *raystartx += step;
            ++strandct;
            if( strandct > this->mMaxStrands )
            {
               break;
            }
         }
         if( strandct > this->mMaxStrands )
         {
            break;
         }

      }
      list_free( &trisurf );
      if( strandct > this->mMaxStrands )
      {
         list_free( &selsurf );
         message_dialog( "Maximum number of strands (%d) exceeded", mMaxStrands );
         return NULL;
      }

   }

   if( mDebugMode )
   {
      printf( "Hair generated for %d surfaces\n", list_count( selsurf ));
      printf( "Creating strands\n" );
   }

   ACObject *o;
   if( this->mStrandSides == 1 )
   {
      if( mDebugMode ) printf( "Creating line strands\n" );
      // Create the strands as line segments
      if( mDebugMode ) printf( "Creating object\n" );
      o = ac_new_object();
      Vertex **varr = new Vertex*[mSegments + 1];
      if( mDebugMode ) printf( "Iterating through strands\n" );
      for( CDStrand *s = mSHead; s != NULL; s = s->mNext )
      {
         // Add the surface
         if( mDebugMode > 12 ) printf( "Creating new surface\n" );
         Surface *surf = new_surface();
         surf->col = 1;
         if( mDebugMode > 12 ) printf( "Adding new surface\n" );
         object_add_surface_head( o, surf );
         // Add the remaining vertices
         for( int ix = 0; ix < mSegments + 1; ++ix )
         {
            if( mDebugMode > 12 ) printf( "Creating vertex\n" );
            varr[ix] = new_vertex_set( s->mSegmentVerts[ix].x, s->mSegmentVerts[ix].y, s->mSegmentVerts[ix].z );
            if( mDebugMode > 12 ) printf( "Adding vertex\n" );
            object_add_vertex_head( o, varr[ix] );
            if( mDebugMode > 12 ) printf( "Adding vertex to surface\n" );
            surface_add_vertex( surf, varr[ix], 0, 0 );
         }
         if( mDebugMode > 12 ) printf( "Setting surface type\n" );
         surface_set_type( surf, SURFACE_LINE );
      }
      if( mDebugMode ) printf( "Adding object to world\n" );
      delete[] varr;
   }
   else
   {
      if( mDebugMode ) printf( "Creating object\n" );
      o = ac_new_object();
      int numverts = ( mSegments + 1 ) * mStrandSides;
      Vertex **varr = new Vertex*[numverts];
      // Extrude strands
      if( mDebugMode ) printf( "Extruding strands\n" );
      float profilerotationangle = 0, lastprofilerotationangle = 0;
      for( CDStrand *s = mSHead; s != NULL; s = s->mNext )
      {
         if( this->mRotateStrands || mStrandSides != 2 )
         {
            // Randomly rotating strands - set arbitrary angle here
            profilerotationangle = DEGTORAD((( float )( rand() % 360 )));
         }
         else
         {
            if( mStrandSides == 2 )
            {
               profilerotationangle = Get2SidedRotationXZ( s, lastprofilerotationangle );
            }
         }
         lastprofilerotationangle = profilerotationangle;
         // Set thickness
         float swidrand = 0.0f;
         if( this->mRandomStrandThickness > CDFUR_EPSILON )
         {
            // Vary thickness from -100% to 100%
            swidrand = (( float )( rand() % 2000 - 1000 )) * 0.001f * this->mRandomStrandThickness * this->mStrandThickness;
         }
         // Extrude
         s->CreateExtrudedStrand( this->mStrandSides, max( 0.001f, this->mStrandThickness + swidrand ), profilerotationangle );
         s->ClearSegments();
         for( int ix = 0; ix < numverts; ++ix )
         {
            if( mDebugMode > 12 ) printf( "Creating vertex\n" );
            varr[ix] = new_vertex_set( s->mExtVerts[ix].x, s->mExtVerts[ix].y, s->mExtVerts[ix].z );
            if( mDebugMode > 12 ) printf( "Adding vertex\n" );
            object_add_vertex_head( o, varr[ix] );
         }
         s->ClearExtruded();
#define MY_SURFACE_ADD_VERTICE(u,v,w,x) surface_add_vertex( surf, varr[(u) * mStrandSides + (v)], (w), (x) )
         // Create strand surface
         // TODO: Adjust if we bring the strand to a point
         float texcoord_inc = 1.0f / (( float )mSegments );
         for( int ix = 0; ix < this->mSegments; ++ix )
         {
            for( int iy = 0; iy < mStrandSides - 1; ++iy )
            {
               // Add the surface
               Surface *surf = new_surface();
               object_add_surface_head( o, surf );
               // Add the vertices
               // NOTE: Texture coordinates will be off if we allow for varying strand lengths
               float tx = (( float )( ix )) * texcoord_inc;
               MY_SURFACE_ADD_VERTICE( ix + 1, iy,     0, tx + texcoord_inc );
               MY_SURFACE_ADD_VERTICE( ix + 1, iy + 1, 1, tx + texcoord_inc );
               MY_SURFACE_ADD_VERTICE( ix,     iy + 1, 1, tx );
               MY_SURFACE_ADD_VERTICE( ix,     iy,     0, tx );
            }
            if( mStrandSides > 2 )
            {
               Surface *surf = new_surface();
               object_add_surface_head( o, surf );
               float tx = (( float )( mSegments - ix ));
               MY_SURFACE_ADD_VERTICE( ix + 1, mStrandSides - 1, 0, tx + texcoord_inc );
               MY_SURFACE_ADD_VERTICE( ix + 1, 0,                1, tx + texcoord_inc );
               MY_SURFACE_ADD_VERTICE( ix,     0,                1, tx );
               MY_SURFACE_ADD_VERTICE( ix,     mStrandSides - 1, 0, tx );
            }
         }
#undef MY_SURFACE_ADD_VERTICE
      }
      if( mDebugMode ) printf( "Adding object to world\n" );
      delete[] varr;
   }

   // Add the object
   object_calc_normals_force( o );
   if( fur_add_undoable )
   {
      add_undoable_world_add_object( "fur", o );
   }
   else
   {
      object_add_child( ac_get_world(), o );
   }

   if( mDebugMode ) printf( "Freeing surface list\n" );
   list_free( &selsurf );

   return o;

}
//---------------------------------------------------------------------------------------
bool CDFur::CheckInputs( bool console_only )
{

   if( mStrandsPerUnit < CDFUR_EPSILON )
   {
      if( console_only )
      {
         printf( "fur : strands per unit is too close to (or less than) zero\n" );
      }
      else
      {
         message_dialog( "fur : strands per unit is too close to (or less than) zero" );
      }
      return false;
   }
   else if( mStrandsPerUnit < CDFUR_EPSILON )
   {
      if( console_only )
      {
         printf( "fur : strand thickness is too close to (or less than) zero\n" );
      }
      else
      {
         message_dialog( "fur : strand thickness is too close to (or less than) zero" );
      }
      return false;
   }
   else if( mStrandsPerUnit > 1.0f / CDFUR_EPSILON )
   {
      if( console_only )
      {
         printf( "fur : too many strands per unit specified\n" );
      }
      else
      {
         message_dialog( "fur : too many strands per unit specified" );
      }
      return false;
   }
   else if( mStrandThickness < CDFUR_EPSILON )
   {
      if( console_only )
      {
         printf( "fur : strand thickness is too close to (or less than) zero\n" );
      }
      else
      {
         message_dialog( "fur : strand thickness is too close to (or less than) zero" );
      }
      return false;
   }
   else if( mSegments < 1 )
   {
      if( console_only )
      {
         printf( "fur : check number of segments per strand\n" );
      }
      else
      {
         message_dialog( "fur : check number of segments per strand" );
      }
      return false;
   }
   else if( mLength < CDFUR_EPSILON )
   {
      if( console_only )
      {
         printf( "fur : strand length is too close to (or less than) zero\n" );
      }
      else
      {
         message_dialog( "fur : strand length is too close to (or less than) zero" );
      }
      return false;
   }
   else if( mStrandSides < 1 )
   {
      if( console_only )
      {
         printf( "fur : number of strand sides is less than or equal to zero\n" );
      }
      else
      {
         message_dialog( "fur : number of strand sides is less than or equal to zero" );
      }
      return false;
   }

   return true;

}
//---------------------------------------------------------------------------------------
void CDFur::DebugPrintSettings()
{
   printf( "\nFur settings\n============\n" );
   printf( "Random placement           : %g\n", mRandomPlacement );
   printf( "Random direction           : %g\n", mRandomDirection );
   printf( "Strands per unit           : %g\n", mStrandsPerUnit );
   printf( "Strand thickness           : %g\n", mStrandThickness );
   printf( "Random strand thickness    : %g\n", mRandomStrandThickness );
   printf( "Segments                   : %d\n", mSegments );
   printf( "Length                     : %g\n", mLength );
   printf( "Random Length              : %g\n", mRandomLength );
   printf( "Rotate Strands             : %s\n", mRotateStrands ? "true" : "false" );
   printf( "Random Seed                : %d\n", mRandomSeed );
   printf( "Gravity                    : %g %g %g\n", mGravity.x, mGravity.y, mGravity.z );
   printf( "Strand Sides               : %d\n", mStrandSides );
   printf( "Maximum hairs              : %d\n", mMaxStrands );
   printf( "\n" );
}
//---------------------------------------------------------------------------------------
float CDFur::Get2SidedRotationXZ( CDStrand *s, float defval )
{

   Vertex up;
   up.x = 0;
   up.y = 1.0f;
   up.z = 0;
   Vertex dirvec;
   Point2 rotvec;
   float rotangle;
   int domdir;

   // Rotate strands in accordance with fur direction/curve
   dirvec.x = fabs( s->mDirection.x );
   dirvec.y = fabs( s->mDirection.y );
   dirvec.z = fabs( s->mDirection.z );

   // Determine the dominant direction axis
   if( dirvec.y >= dirvec.x && dirvec.y >= dirvec.z )
   {
      domdir = 1;
   }
   else if( dirvec.x >= dirvec.y && dirvec.x >= dirvec.z )
   {
      domdir = 0;
   }
   else if( dirvec.z >= dirvec.x && dirvec.z >= dirvec.y )
   {
      domdir = 2;
   }

   // Set base rotation based on strand direction
   switch( domdir )
   {

   // X axis
   case 0:
      rotvec.x = s->mSegmentVerts[1].y - s->mSegmentVerts[0].y;
      rotvec.y = s->mSegmentVerts[1].z - s->mSegmentVerts[0].z;
      if( rotvec.x == 0 && rotvec.y == 0 )
      {
         // Vector points along axis - use the default
         rotangle = defval;
      }
      else
      {
         float veclen = sqrtf( pow( rotvec.x, 2 ) + pow( rotvec.y, 2 ));
         if( veclen == 0 )
         {
            rotangle = defval;
         }
         else
         {
            rotvec.x /= veclen;
            rotvec.y /= veclen;
            rotangle = acos( rotvec.x * 0.0f + rotvec.y * 1.0f );
         }
      }
      break;

   // Z axis
   case 2:
      rotvec.x = s->mSegmentVerts[1].x - s->mSegmentVerts[0].x;
      rotvec.y = s->mSegmentVerts[1].y - s->mSegmentVerts[0].y;
      if( rotvec.x == 0 && rotvec.y == 0 )
      {
         // Vector points along axis - use the default
         rotangle = defval;
      }
      else
      {
         float veclen = sqrtf( pow( rotvec.x, 2 ) + pow( rotvec.y, 2 ));
         if( veclen == 0 )
         {
            rotangle = defval;
         }
         else
         {
            rotvec.x /= veclen;
            rotvec.y /= veclen;
            rotangle = acos( rotvec.x * 1.0f + rotvec.y * 0.0f ) - DEGTORAD( 90.0f );
         }
      }
      break;

   // Y axis
   default:
      rotvec.x = s->mSegmentVerts[1].x - s->mSegmentVerts[0].x;
      rotvec.y = s->mSegmentVerts[1].z - s->mSegmentVerts[0].z;
      if( rotvec.x == 0 && rotvec.y == 0 )
      {
         // Vector points along axis - use the default
         rotangle = defval;
      }
      else
      {
         float veclen = sqrtf( pow( rotvec.x, 2 ) + pow( rotvec.y, 2 ));
         if( veclen == 0 )
         {
            rotangle = defval;
         }
         else
         {
            rotvec.x /= veclen;
            rotvec.y /= veclen;
            if( rotvec.x < 0.0f )
            {
               rotvec.y = -( rotvec.y );
            }
            rotangle = acos( rotvec.x * 0.0f + rotvec.y * (-1.0f) );
         }
      }
      break;

   }

   return rotangle;

}
//---------------------------------------------------------------------------------------
CDStrand *CDFur::FindDuplicateStrand( Vertex *v1, Vertex *v2, Vertex *v3, Vertex *strandloc )
{
   // Gather list of adjacent surfaces
   List *surflist = NULL;
   List *lt = NULL;
   lt = surfaces_of_object_vertex( object_of_vertex( v1 ), v1 );
   list_append_list( &surflist, lt );
//   list_free( &lt );
   lt = surfaces_of_object_vertex( object_of_vertex( v2 ), v2 );
   list_append_list_unique( &surflist, lt );
//   list_free( &lt );
   lt = surfaces_of_object_vertex( object_of_vertex( v3 ), v3 );
   list_append_list_unique( &surflist, lt );
//   list_free( &lt );
   // Search for dupes
   for( CDStrand *str = mSHead; str != NULL; str = str->mNext )
   {
      for( List *s = surflist; s != NULL; s = s->next )
      {
         Surface *surf = ( Surface * )( s->data );
         if( surf != str->mParentSurface )
         {
            continue;
         }
         Vertex *vt1 = &( str->mRoot );
         if( fabs( vt1->x - strandloc->x ) < CDFUR_EPSILON &&
             fabs( vt1->y - strandloc->y ) < CDFUR_EPSILON &&
             fabs( vt1->z - strandloc->z ) < CDFUR_EPSILON )
         {
            return str;
         }
      }
   }
   list_free( &surflist );
   return NULL;
}
//---------------------------------------------------------------------------------------
