#include "cdstrand.h"
#include "cdstrandprofile.h"
#include <math.h>
#include <float.h>
#include <stdio.h>

//---------------------------------------------------------------------------------------
CDStrand::CDStrand( unsigned short numsegments )
{
   mParentSurface = NULL;
   mNumSegments = numsegments;
   mSegmentVerts = new Vertex[mNumSegments + 1];
   mSegmentDirs = new Vertex[mNumSegments + 1];
   mNext = NULL;
   mExtVerts = NULL;
   mNumExtVerts = 0;
   mDebug = 0;
}
//---------------------------------------------------------------------------------------
CDStrand::~CDStrand()
{
   ClearSegments();
   ClearExtruded();
}
//---------------------------------------------------------------------------------------
void CDStrand::ClearSegments()
{
   if( mSegmentVerts != NULL )
   {
      delete[] mSegmentVerts;
      mSegmentVerts = NULL;
   }
   if( mSegmentDirs == NULL )
   {
      delete[] mSegmentDirs;
      mSegmentDirs = NULL;
   }
   mNumSegments = 0;
}
//---------------------------------------------------------------------------------------
void CDStrand::ClearExtruded()
{
   if( mExtVerts != NULL )
   {
      delete[] mExtVerts;
      mExtVerts = NULL;
   }
   mNumExtVerts = 0;
}
//---------------------------------------------------------------------------------------
void CDStrand::CreateExtrudedStrand( int numverts, float diameter, float rotprofangle )
{

   // This code is a re-interpretation of Andy Thaller's "Extrude Along Path" code for AC3D.
   // It essentially *is* the code, but rearranged quite a bit to work with
   // constructs in the fur routine.
   // See header file for more licensing/credit.

   // Allocate memory to the extruded strand
   // NOTE: if we let the extruded version taper to zero, account for this here
   // TODO: Should create this outside of strand routine, since the same profile will be used for all hairs.
   ClearExtruded();
   mNumExtVerts = numverts * ( mNumSegments + 1 );
   mExtVerts = new Vertex[mNumExtVerts];

   // Create the strand profile
   CDStrandProfile profile( numverts );
   profile.CreateStrandProfileXZ( diameter, rotprofangle );
   if( mDebug > 12 )
   {
      printf( "strand : creating profile:\n" );
      for( int ix = 0; ix < numverts; ++ix )
      {
         printf( "strand :    %g %g %g\n", profile.mVerts[ix].x, profile.mVerts[ix].y, profile.mVerts[ix].z );
      }
   }

   // Additional strand profiles are used to move and rotate the original
   // strand profile.  Lines are cast between the translated/rotated profiles to
   // determine adjoining vertex intersection points.
   CDStrandProfile p_base1( numverts );
   CDStrandProfile p_base2( numverts );
   CDStrandProfile p_dest1( numverts );
   CDStrandProfile p_dest2( numverts );
   CDStrandProfile p_store( numverts );

   // Vector direction
   Vertex *dirvec1, *dirvec2;
   // Vector location
   Vertex *locvec;
   // Distance between location vectors
   Vertex locvec_dist;

   Vertex axis;
   double cosa;

   int vertctr = 0;

   //////////////////////////////////////////////////////////////////////////
   // Create the extrusion base

   if( mDebug > 12 ) printf( "strand : === CREATING STRAND BASE ===\n" );

   // Grab the first direction/location vector
   dirvec1 = &( mSegmentDirs[0] );
   locvec = &( mSegmentVerts[0] );
   if( mDebug > 12 ) printf( "strand : direction 1 : %g %g %g\n", dirvec1->x, dirvec1->y, dirvec1->z );
   if( mDebug > 12 ) printf( "strand : location 1 : %g %g %g\n", locvec->x, locvec->y, locvec->z );

   // If the y direction of the first point of the path is negative (away), reverse the direction
   // of our profile object.
   // If we didn't know our profile's normal already, this would be more involved.
   Vertex norm;
   norm.x = 0;
   norm.z = 0;
   if( dirvec1->y < 0.0 )
   {
      norm.y = -1;
      profile.ReverseVertexOrder();
   }
   else
   {
      norm.y = 1;
   }
   if( mDebug > 12 ) printf( "strand : normal : %g %g %g\n", norm.x, norm.y, norm.z );
   // This is the new direction of our base profile object.
   cosa = C3d::VectorCosAngle( &norm, dirvec1 );
   if( mDebug > 12 ) printf( "strand : cosangle : %g\n", cosa );

   // Populate the rotation profile
   p_base1.CopyFrom( &profile );

   // Set the initial p_base1 position
   p_base1.MoveBy( locvec );

   // Check whether template and first path vertex are already aligned
   if( cosa == 1.0 || !_finite( cosa ))
   {
      // No action required for first profile
      if( mDebug > 12 ) printf( "strand : no rotation of base angle required\n" );
   }
   else
   {
      // Rotate profile
      C3d::VectorAxis( &norm, dirvec1, &axis );
      if( mDebug > 12 ) printf( "strand : axis : %g %g %g\n", axis.x, axis.y, axis.z );
      p_base1.RotateAxis( locvec, &axis, cosa );
   }

   // Add the vertices in the base profile
   memcpy( &( mExtVerts[vertctr] ), p_base1.mVerts, sizeof( Vertex ) * numverts );
   if( mDebug > 12 )
   {
      printf( "strand : Vertices added to base profile\n" );
      if( mDebug > 12 )
      {
         printf( "strand : rotated profile:\n" );
         for( int ix = vertctr; ix < vertctr + numverts; ++ix )
         {
            printf( "strand :    %g %g %g\n", mExtVerts[ix].x, mExtVerts[ix].y, mExtVerts[ix].z );
         }
      }
   }
   vertctr += numverts;

   // Move p_base2 to the next position
   locvec_dist.x = mSegmentVerts[1].x - locvec->x;
   locvec_dist.y = mSegmentVerts[1].y - locvec->y;
   locvec_dist.z = mSegmentVerts[1].z - locvec->z;
   p_base2.CopyFrom( &p_base1 );
   p_base2.MoveBy( &locvec_dist );

   // If there is only 1 segment, we're done
   if( mNumSegments == 1 )
   {
      if( mDebug > 12 ) printf( "strand : === PROFILE 2 ===\n" );
      memcpy( &( mExtVerts[vertctr] ), p_base2.mVerts, sizeof( Vertex ) * numverts );
      if( mDebug > 12 )
      {
         printf( "strand : extruded profile:\n" );
         for( int ix = vertctr; ix < vertctr + numverts; ++ix )
         {
            printf( "strand :    %g %g %g\n", mExtVerts[ix].x, mExtVerts[ix].y, mExtVerts[ix].z );
         }
      }
      vertctr += numverts;
   }

   // If more than 1 segment, add remaining
   else
   {

      // Extrude/add remaining profiles
      for( int ix = 2; ix < this->mNumSegments + 1; ++ix )
      {

         if( mDebug > 12 ) printf( "strand : === PROFILE %d ===\n", ix );
         dirvec2 = &( mSegmentDirs[ix-1] );
         locvec = &( mSegmentVerts[ix-1] );
         if( mDebug > 12 ) printf( "strand : direction : %g %g %g\n", dirvec2->x, dirvec2->y, dirvec2->z );
         if( mDebug > 12 ) printf( "strand : location : %g %g %g\n", locvec->x, locvec->y, locvec->z );

         // This is the new direction of our profile object.
         cosa = C3d::VectorCosAngle( dirvec1, dirvec2 );
         if( mDebug > 12 ) printf( "strand : cosangle : %g\n", cosa );

         locvec_dist.x = mSegmentVerts[ix].x - locvec->x;
         locvec_dist.y = mSegmentVerts[ix].y - locvec->y;
         locvec_dist.z = mSegmentVerts[ix].z - locvec->z;
         if( mDebug > 12 ) printf( "strand : locdist: %g %g %g\n", locvec_dist.x, locvec_dist.y, locvec_dist.z );

         // Check whether template and first path vertex are already aligned
         if( cosa == 1.0 || !_finite( cosa ))
         {
            // Same angle
            if( mDebug > 12 ) printf( "strand : no rotation necessary\n" );
            p_dest1.CopyFrom( &p_base2 );
            p_dest2.CopyFrom( &p_dest1 );
            p_dest2.MoveBy( &locvec_dist );
            memcpy( &( mExtVerts[vertctr] ), p_dest1.mVerts, sizeof( Vertex ) * numverts );
            if( mDebug > 12 )
            {
               printf( "strand : extruded verts:\n" );
               for( int iw = vertctr; iw < vertctr + numverts; ++iw )
               {
                  printf( "strand :    %g %g %g\n", mExtVerts[iw].x, mExtVerts[iw].y, mExtVerts[iw].z );
               }
            }
            vertctr += numverts;
         }
         else
         {
            // Rotate profile
            C3d::VectorAxis( dirvec1, dirvec2, &axis );
            if( mDebug > 12 ) printf( "strand : rot axis : %g %g %g\n", axis.x, axis.y, axis.z );
            p_dest1.CopyFrom( &p_base2 );
            p_dest1.RotateAxis( locvec, &axis, cosa );
            if( mDebug > 12 )
            {
               printf( "strand : rotated profile\n" );
               for( int iw = 0; iw < numverts; ++iw )
               {
                  printf( "strand :    %g %g %g\n", p_dest1.mVerts[iw].x, p_dest1.mVerts[iw].y, p_dest1.mVerts[iw].z );
               }
            }
            p_dest2.CopyFrom( &p_dest1 );
            p_dest2.MoveBy( &locvec_dist );
            // Get the intersection between lines into store
            p_store.GetIntersections( &p_base1, &p_base2, &p_dest1, &p_dest2 );
            if( mDebug > 12 )
            {
               printf( "strand : intersection verts:\n" );
               for( int iw = 0; iw < numverts; ++iw )
               {
                  printf( "strand :    %g %g %g\n", p_store.mVerts[iw].x, p_store.mVerts[iw].y, p_store.mVerts[iw].z );
               }
            }
            memcpy( &( mExtVerts[vertctr] ), p_store.mVerts, sizeof( Vertex ) * numverts );
            if( mDebug > 12 )
            {
               printf( "strand : extruded verts:\n" );
               for( int iw = vertctr; iw < vertctr + numverts; ++iw )
               {
                  printf( "strand :    %g %g %g\n", mExtVerts[iw].x, mExtVerts[iw].y, mExtVerts[iw].z );
               }
            }
            vertctr += numverts;
         }

         // Prepare for next iteration
         if( mDebug > 12 ) printf( "strand : \n" );
         dirvec1 = dirvec2;
         p_base1.CopyFrom( &p_dest1 );
         p_base2.CopyFrom( &p_dest2 );

      }

      // Add the final profile
      memcpy( &( mExtVerts[vertctr] ), p_dest2.mVerts, sizeof( Vertex ) * numverts );
      vertctr += numverts;

   }

}
//---------------------------------------------------------------------------------------
