#include "cdfur.h"
#include <stdio.h>
#include <string.h>

static ACObject *g_DDHFurCreatedObject = NULL;
static bool g_DDHFurStoreSettings = false;

// Some functions (noted in comments within each source code header, where appropriate)
//    were strongly derived or outright copied from Andy Thaller's Extrude Along Path plugin for AC3D.
// Such code was used in accordance with the GNU General Public License.
// For more information on Andy Thaller's work, see http://www.ph.tum.de/~thaller/extrude/.

//---------------------------------------------------------------------------------------
// Disable realtime updates
static void ddhfur_disable_realtime()
{
   tcl_command( "grab release .ddhfur" );
   //printf( "fur : realtime updates are off\n" );
   // Remove the old hair object
   if( object_exists( g_DDHFurCreatedObject ))
   {
      object_delete( g_DDHFurCreatedObject );
      redraw_all();
      display_status();
   }
   g_DDHFurCreatedObject = NULL;
}
//---------------------------------------------------------------------------------------
// Enable realtime updates
static void ddhfur_enable_realtime()
{
   tcl_command( "grab .ddhfur" );
   //printf( "fur : realtime updates are on\n" );
}
//---------------------------------------------------------------------------------------
static void ddhfur_set_save_options( int argc, const char* argv[] )
{
   if( argc == 2 )
   {
      if( argv[1][0] == '0' && argv[1][1] == '\0' )
      {
         g_DDHFurStoreSettings = false;
      }
      else
      {
         g_DDHFurStoreSettings = true;
      }
   }
}
//---------------------------------------------------------------------------------------
// Create hair
static void ddhfur_go( int argc, const char* argv[] )
{

   if( argc < 20 )
   {
      message_dialog( "Insufficient parameters passed to fur plugin" );
      return;
   }

   // Fur object
   CDFur fur;

   // Get params
   if( sscanf( argv[1], "%g", &( fur.mStrandsPerUnit )) != 1 )
   {
      message_dialog( "Error parsing strands per unit (%s)", argv[1] );
      return;
   }
   if( sscanf( argv[2], "%g", &( fur.mRandomPlacement )) != 1 )
   {
      message_dialog( "Error parsing % random placement (%s)", argv[2] );
      return;
   }
   fur.mSegments = atoi( argv[3] );
   if( sscanf( argv[4], "%g", &( fur.mLength )) != 1 )
   {
      message_dialog( "Error parsing strand length (%s)", argv[4] );
      return;
   }
   if( sscanf( argv[5], "%g", &( fur.mRandomLength )) != 1 )
   {
      message_dialog( "Error parsing % random length (%s)", argv[5] );
      return;
   }
   if( sscanf( argv[6], "%g", &( fur.mStrandThickness )) != 1 )
   {
      message_dialog( "Error parsing strand thickness (%s)", argv[6] );
      return;
   }
   if( sscanf( argv[7], "%g", &( fur.mRandomStrandThickness )) != 1 )
   {
      message_dialog( "Error parsing % random thickness (%s)", argv[7] );
      return;
   }
   fur.mStrandSides = atoi( argv[8] );
   fur.mRotateStrands = ( atoi( argv[9] ) != 0 );
   if( sscanf( argv[10], "%g", &( fur.mRandomDirection )) != 1 )
   {
      message_dialog( "Error parsing % random direction (%s)", argv[10] );
      return;
   }
   if( sscanf( argv[11], "%g", &( fur.mGravity.x )) != 1 )
   {
      message_dialog( "Error parsing X-axis gravity (%s)", argv[11] );
      return;
   }
   if( sscanf( argv[12], "%g", &( fur.mGravity.y )) != 1 )
   {
      message_dialog( "Error parsing Y-axis gravity (%s)", argv[12] );
      return;
   }
   if( sscanf( argv[13], "%g", &( fur.mGravity.z )) != 1 )
   {
      message_dialog( "Error parsing Z-axis gravity (%s)", argv[13] );
      return;
   }
   if( sscanf( argv[14], "%d", &( fur.mMaxStrands )) != 1 )
   {
      message_dialog( "Error parsing maximum strands (%s)", argv[14] );
      return;
   }
   fur.mMaxStrands = ( unsigned int )abs(( double )fur.mMaxStrands );
   if( sscanf( argv[15], "%g", &( fur.mRootDepth )) != 1 )
   {
      message_dialog( "Error parsing root depth (%s)", argv[15] );
      return;
   }
   bool realtime_on = ( strcmp( "0", argv[16] ) != 0 );
   bool preview_only = ( strcmp( "0", argv[17] ) != 0 );
   int randseed = 0;
   if( sscanf( argv[18], "%d", &randseed ) != 1 )
   {
      message_dialog( "Error parsing random seed (%s)", argv[18] );
      return;
   }
   if( randseed != 0 )
   {
      fur.mRandomSeed = randseed;
   }
   fur.mRemoveDuplicates = ( strcmp( "0", argv[19] ) != 0 );

   // Remove the old hair object
   if( object_exists( g_DDHFurCreatedObject ))
   {
      object_delete( g_DDHFurCreatedObject );
   }
   g_DDHFurCreatedObject = NULL;

   // Only want the number of strands in one direction, not the whole unit
   fur.mStrandsPerUnit = sqrtf( fur.mStrandsPerUnit );
   // Convert to a percentage of one
   fur.mRandomPlacement *= 0.01f;
   fur.mRandomLength *= 0.01f;
   fur.mRandomStrandThickness *= 0.01f;
   fur.mRandomDirection *= 0.01f;
   fur.mGravity.x *= -( 0.01f );
   fur.mGravity.y *= -( 0.01f );
   fur.mGravity.z *= -( 0.01f );
   // abs randoms
   fur.mRandomPlacement = fabs( fur.mRandomPlacement );
   fur.mRandomLength = fabs( fur.mRandomLength );
   fur.mRandomStrandThickness = fabs( fur.mRandomStrandThickness );
   fur.mRandomDirection = fabs( fur.mRandomDirection );

   if( !fur.CheckInputs( preview_only ) )
   {
      // Failed param verification
      return;
   }

   ACObject *o = fur.Generate( !preview_only );
   if( o != NULL )
   {
      if( realtime_on )
      {
         g_DDHFurCreatedObject = o;
         if( !preview_only )
         {
            ddhfur_disable_realtime();
         }
      }
   }

   // Write the fur settings
   if(( !preview_only ) && ( g_DDHFurStoreSettings ))
   {
      char furset[256];
      char furtex[16384];
      strcpy( furtex, "// Fur settings (1.0.1)" );
      sprintf( furset, "\n// Strands per unit: %s", argv[1] );
      strcat( furtex, furset );
      sprintf( furset, "\n// %% random placement: %s", argv[2] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Segments per strand: %s", argv[3] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Strand length: %s", argv[4] );
      strcat( furtex, furset );
      sprintf( furset, "\n// %% random length: %s", argv[5] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Root depth: %s", argv[15] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Maximum strands: %s", argv[14] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Remove duplicates: %s", fur.mRemoveDuplicates ? "on" : "off" );
      strcat( furtex, furset );
      sprintf( furset, "\n// Thickness: %s", argv[6] );
      strcat( furtex, furset );
      sprintf( furset, "\n// %% random thickness: %s", argv[7] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Sides per strand: %s", argv[8] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Randomly rotate strands: %s", fur.mRotateStrands ? "on" : "off" );
      strcat( furtex, furset );
      sprintf( furset, "\n// %% random direction: %s", argv[10] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Gravity X: %s", argv[11] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Gravity Y: %s", argv[12] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Gravity Z: %s", argv[13] );
      strcat( furtex, furset );
      sprintf( furset, "\n// Random seed: %d", fur.mRandomSeed );
      strcat( furtex, furset );
      object_set_userdata( o, furtex );
   }

   if( fur.mDebugMode ) printf( "Redrawing\n" );
   redraw_all();
   if( fur.mDebugMode ) printf( "Displaying status\n" );
   display_status();
   if( fur.mDebugMode ) printf( "Done\n" );

}
//---------------------------------------------------------------------------------------
// Realtime updating
static void ddhfur_sync_realtime( int argc, const char* argv[] )
{

   if( argc < 2 )
   {
      return;
   }

   bool realtime_on = ( strcmp( "0", argv[1] ) != 0 );

   // Realtime updates are off
   if( realtime_on )
   {
      if( nothing_is_selected() )
      {
         tcl_command( ".ddhfur.frmrealtime.ddhfur_realtime invoke" );
      }
      else
      {
         ddhfur_enable_realtime();
      }
   }
   else
   {
      ddhfur_disable_realtime();
   }

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      g_DDHFurCreatedObject = NULL;

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/fur.tcl\"", path );
      ac_add_tools_menu_item( "Fur...", "ddhfur_dialog",
                              "Apply strands of fur to the selected surfaces" );
      ac_add_command_full( "ddhfur_go", ( void * )ddhfur_go,
                            0, "argv", "<arg1> <arg2>",
                           "Create fur" );
      ac_add_command_full( "ddhfur_sync_realtime", ( void * )ddhfur_sync_realtime,
                            0, "argv", "<arg1>",
                           "Sync realtime fur interface" );
      ac_add_command_full( "ddhfur_set_save_options", ( void * )ddhfur_set_save_options,
                            0, "argv", "<arg1>",
                           "Sets whether to save generated fur settings to the object's Data field" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Fur - version 1.0 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Fur - Dennis Hawthorne - Applies strands of fur to selected surfaces.\nPolygonal hair extrusions derived from Andy Thaller's Extrude Along Path plugin." );
}
//---------------------------------------------------------------------------------------
